import editdistance
# pip install editdistance
# or https://pypi.python.org/pypi/editdistance/0.2

import argparse
import os, sys, time 
from collections import Counter
from itertools import count
from operator import add

# import nltk.corpus
# We hardcoded the stopwords list we needed from nltk to not depend on it.

parser = argparse.ArgumentParser()
parser.add_argument("--collection_tuples", "-i", type=str, dest="collection_path",
                    help="the extractions database")
parser.add_argument("--questions_reference", "-r", type=str, dest="qa_path",
                    help="the QAs reference file")

parser.add_argument("--system_name", "-s", type=str, dest="system", default="unknown",
                    help="the name of the system evaluated")
parser.add_argument("--verbose", "-v", type=str, dest="verbose", default=False,
                    help="print much details")
parser.add_argument("--output_flag", "-f", type=str, dest="paper_flag", default=None,
                    help="flag to set if we want to build the paper tables of results. Values it can take are \"parameter-table\" and \"perf-table\".")


parser.add_argument("--editdistance", "-e", type=int, dest="word_matching_editdistance_threshold", default=1,
                    help="two words are a match if less than # of their chars mismatch")
parser.add_argument("--match_question_threshold", "-m", type=float, dest="question_fact_matching_threshold", default=0.6,
                    help="an extracted fact answers a question if it matches its words more than this")
# This default value is an educated guess.
# 0.7 also makes sense. 0.8 seems too restrictive.
# 0.5 is too loose (~25 answers per question instead of <5)
# 0.6 works well.

args = parser.parse_args()

print >> sys.stderr, time.strftime('%X'), "- Parameters: an edit distance of", args.word_matching_editdistance_threshold, "is allowed between question and extraction words to match, and a fact answers question if it matches more than" , "{:.0%}".format(args.question_fact_matching_threshold), "of its nonstopwords."


if args.verbose:
    verbose = open("../logs/"+time.strftime('%Y-%m-%d-%H:%M:%S')+"__"+args.system+"_qa-perf.verbose_log.txt", "w")
else:
    verbose = open(os.devnull, "w")
    
print >> verbose, "Verbose log for the run of the system :", args.system, "running on", time.strftime('%Y-%m-%d-%H:%M:%S')
print >> verbose, "Tuple collection input was at", args.collection_path
print >> verbose, "Q&A file was at", args.qa_path
print >> verbose, ""

def avg(l):
    try:
        return sum(l)/len(l)
    except ZeroDivisionError:
        return float("inf")

class Extraction:
    def __init__(self, parts):
        self.arg1 = parts[2]
        self.rel = parts[3]
        self.arg2 = parts[4]
    def text_form (self):
        return " ".join(self.parts())
    def parts(self):
        return([self.arg1, self.rel, self.arg2])

class QA:
    def __init__(self):
        self.question = None
        self.answer = None
        self.answers = []
        self.sentence = None
        self.qtype = None
        self.expectation = None

qas = []
# Parse questions
with open(args.qa_path) as qa_file:
    qa = QA()
    for line in qa_file.readlines():
        if line == "\n":
            qa = QA()
        elif "S:" == line[:2]:
            qa.sentence = line[2+1:-1] # +1 is a space
        elif "Q:" in line and line[0] != "#":
            qa.question = line[line.index(":")+1:-1]
            qtype = line[:line.index(":")-1] # -1 is the 'Q'
            qa.qtype = qtype if qtype == "Y" else "F" # YQ for Yes/No or Q for Factoid (default)
            qas.append(qa)
        elif "A:" == line[:2]:
            qa.answer = line[3:-1]
        elif "X:" == line[:2]:
            qa.expectation = int(line[3:-1])
        # We don't read the other lines

# These evaluation bricks we can compute before actually running anything.
n_expect_q = sum(1 for qa in qas if qa.expectation is not None) # Is 0 or 1
n_yes_q = sum(1 for qa in qas if qa.qtype == "Y" and qa.answer == "Yes")
n_no_q = sum(1 for qa in qas if qa.qtype == "Y" and qa.answer == "No")
n_yn_qas = sum(1 for qa in qas if qa.qtype == "Y")
n_fact_q = sum(1 for qa in qas if qa.qtype == "F")
n_yn_q = n_no_q + n_yes_q
n_q = n_fact_q + n_yn_q
assert (n_q == len(qas)) # sanity check

print >> sys.stderr, time.strftime('%X'), "- Imported Q&A:", len(qas), "pairs, at", args.qa_path
print >> sys.stderr, n_expect_q, "questions have an expected system-performance score."
print >> sys.stderr, n_yn_qas, "are Yes/No questions ("+str(n_yes_q)+"/"+str(n_no_q)+") and", n_fact_q, "require a factoid answer."

# Load collection
collection_tuples = []

print >> sys.stderr, time.strftime('%X'), "- Counting candidate facts at", args.collection_path
num_exts = sum(1 if line[:3]=="EXT" else 0 for line in open(args.collection_path))
print >> sys.stderr, time.strftime('%X'), "- Measuring coverage of", num_exts, "extractions on", len(qas), "questions."

with open(args.collection_path, "r") as sef:
    stored_exts = []
    for line in sef:
        parts = line.decode("utf-8")[:-1].split("\t")
        if parts[0][:-1] == "EXT":
            ext = Extraction(parts) if len(parts)>4 else None
            # This discards Clausie's unary extractions
            if ext:
                collection_tuples.append(ext)
# Returns collection_tuples

def content_words(string, stopwords):
    words = string.split(" ")
    content_words = [w for w in words if w not in stopwords]
    return content_words

def fact_question_matcher(question_cwords, fact):
    """Measures how well a fact matches a question."""
    fact_words = fact.text_form().split(" ")
    matching_words = 0.0
    # Does the tuple match the question ?
    for w in question_cwords:
        matching_words += float(w in fact_words or min([editdistance.eval(w, f) for f in fact_words]) <= args.word_matching_editdistance_threshold )
        # lemmatizing would help this part

    matching_score = matching_words / len(question_cwords)
    return matching_score

def extract_answer(question_cwords, fact):
    """Given a factoid Q and a fact answering it, extract the answer."""
    # To -  What was Mrs. Yogeswaran shot with ?
    # The words of which to be matched being -  What Mrs. Yogeswaran shot ?
    # How is the following tuple matching ?
    # EXTA	0.500	Mrs. Yogeswaran	; was shot  with ;	a pistol

    # For an answer, find the part of the tuple that less matched the question.
    parts_in_words = [p.split() for p in fact.parts()]
    tuple_matches = [0,0,0] # number of Q-words that matched each part
    for w in question_cwords:
        # find the tuple word that matches best (min(i, key = Levenstein(w, tw[i])) i.e. for i,tw in enumerate(tuple_words))
        tuple_matches = [sum(x) for x in zip(tuple_matches, [(w in part or min([editdistance.eval(w, tw) for tw in part]) <= args.word_matching_editdistance_threshold) for part in parts_in_words])]

    if tuple_matches[2] == 0:
        # In the most normal case that the second argument was not mentionned by the question, answer it
        answer = fact.arg2
    else:
        # Else answer the tuple part that least matched the question
        answer = fact.parts()[tuple_matches.index(min(tuple_matches))]
    # answer = fact.arg2 if not answer else answer
    # In case it's still undefined, return the second argument, shouldn't happen.

    if "Yogeswaran" in question_cwords:
        print >> verbose, "\ndebug run : extract_answer - Yogeswaran"
        print >> verbose, "fact is :", " \t".join(fact.parts()).encode('utf8')
        print >> verbose, "question matching words are :", question_cwords
        print >> verbose, "tuple_matches :", tuple_matches
        print >> verbose, "answer index :", tuple_matches.index(min(tuple_matches))
        print >> verbose, "returned answer :", answer.encode('utf8')

    return answer

# Now we evaluate the extractions against the QAs.
def valid_answer(answer, qa):
    """What is our criterion for a good answer ?
    For now, if all the gold answer words are in the response, it's good. It's favored to give longer answers."""
    if qa.qtype == "Y":
        return answer == qa.answer
    elif qa.qtype == "F":
        return answer is not None and all(w in answer for w in qa.answer.split(" "))
    else:
        raise NameError(qa.qtype + " : What is this question's type ? "+ qa.question)

mycustomstopwords = ["", "?", ".", " ", "'s"]

# The commented line is ofc better than hardcoding values, but let's not depend on NLTK just for that list of words.
# nltk_stopwords = nltk.corpus.stopwords.words('english')
nltk_stopwords = [u'i', u'me', u'my', u'myself', u'we', u'our', u'ours',
                  u'ourselves', u'you', u'your', u'yours', u'yourself', u'yourselves', u'he',
                  u'him', u'his', u'himself', u'she', u'her', u'hers', u'herself', u'it',
                  u'its', u'itself', u'they', u'them', u'their', u'theirs', u'themselves',
                  u'what', u'which', u'who', u'whom', u'this', u'that', u'these', u'those',
                  u'am', u'is', u'are', u'was', u'were', u'be', u'been', u'being', u'have',
                  u'has', u'had', u'having', u'do', u'does', u'did', u'doing', u'a', u'an',
                  u'the', u'and', u'but', u'if', u'or', u'because', u'as', u'until', u'while',
                  u'of', u'at', u'by', u'for', u'with', u'about', u'against', u'between',
                  u'into', u'through', u'during', u'before', u'after', u'above', u'below',
                  u'to', u'from', u'up', u'down', u'in', u'out', u'on', u'off', u'over',
                  u'under', u'again', u'further', u'then', u'once', u'here', u'there', u'when',
                  u'where', u'why', u'how', u'all', u'any', u'both', u'each', u'few', u'more',
                  u'most', u'other', u'some', u'such', u'no', u'nor', u'not', u'only', u'own',
                  u'same', u'so', u'than', u'too', u'very', u's', u't', u'can', u'will',
                  u'just', u'don', u'should', u'now']

mystopwords = nltk_stopwords + mycustomstopwords

# Init evaluation metrics
n_yes_q_answed_well = n_yes_ans = n_no_q_answed_well = 0
n_fact_q_answed = n_fact_q_answed_well = n_fact_q_noans = 0
n_expect_q_valid = 0
n_fact_ans_in_list = 0
total_num_answers = 0
for i, pair in enumerate(qas):
    print >> sys.stderr, "\r", time.strftime('%X'), "- Evaluating question #", i+1, "/", len(qas),
    question_type = pair.qtype
    question_content_words = content_words(pair.question, mystopwords)
    if question_content_words == None:
        print >> sys.stderr, "\nPb: No content words in this question :"
        print >> sys.stderr, "sentence:", pair.sentence.encode('utf8')
        print >> sys.stderr, "question:", pair.question.encode('utf8')
        print >> sys.stderr, "answer:", pair.answer.encode('utf8')

    # Fetch system answers
    system_answers = {} # dict of (answers : [best score, corresponding fact])
    for fact in collection_tuples:
        # Is_there_an_answer_in_there_?(question_content_words, fact)
        matching_score = fact_question_matcher(question_content_words, fact)

        # Do many enough question words fit the tuple ? If so, it's a match.
        if matching_score >= args.question_fact_matching_threshold:
            if question_type == "Y":
                # Future work : We could detect if there's a match containing a negation, and answer No in this case.
                # For now, we answer No iff no tuple whatsoever match the question.
                system_answers["Yes"] = [matching_score, fact]
                break # we got a yes, we're done
            elif question_type == "F": # (it's really an else)
                answer = extract_answer(question_content_words, fact)
                if answer in system_answers:
                    if system_answers[answer][0] < matching_score:
                        # replace the existing answer iff we improve on it
                        system_answers[answer] = [matching_score, fact]
                else: # it's new answer
                    system_answers[answer] = [matching_score, fact]

    # We passed through all facts and collected all answers along with their support.
    print >> verbose, "\nTo -", pair.question.encode('utf8')
    print >> verbose, "The words of which to be matched being -", " ".join(content_words(pair.question, mystopwords)).encode('utf8')
    print >> verbose, "The answer of which is -", pair.answer.encode('utf8')
    print >> verbose, "The system finds these", len(system_answers.keys()), "as possible answer/tuples :", "\n" + "\n".join(
        ["{:.2}".format(fact_question_matcher(content_words(pair.question, mystopwords), support[1])) + ":"+ a.encode('utf8') +"\n"+"{:.2}".format(support[0])+":"+ support[1].text_form().encode('utf8')
         for a, support in system_answers.items()])
    

    # Now put the bag of answers in order.
    correct_answer_in_the_list = False # Stays false for Y/N questions, only used for recall of factoid ones.
    if question_type == "Y":
        if system_answers.keys():
            answer = "Yes"
            answer_score, support_fact = system_answers[answer]
        else:
            answer = "No"
    else: # only keep top answers.
        if system_answers.keys():
            answer = sorted(system_answers.keys(), key = lambda x: system_answers[x][0], reverse = True)[0]
            answer_score, support_fact = system_answers[answer]
            print >> verbose, "The system final answer/tuple -", answer.encode('utf8'), "\nScore", answer_score, "/ support -", support_fact.text_form().encode('utf8')

            correct_answer_in_the_list = any(valid_answer(answer, pair) for answer in system_answers.keys())
            numof_answers = len(system_answers)
            total_num_answers += numof_answers
            print >> verbose, "Some" if correct_answer_in_the_list else "No", "correct answer was in the full list of answers."
        else:
            answer = None
            correct_answer_in_the_list = False
            print >> verbose, "No answer was found for this question."

    # Now we have one single answer to each question, or None for factoids.
    # We also got answer_score and support_fact (a YaTuple / support_fact.text_form())

    # Performance measurements time
    if valid_answer(answer, pair):
        if pair.qtype == "F" :
            n_fact_q_answed += 1
            n_fact_q_answed_well += 1
            print >> verbose, "Factoid question is well answered."
        elif pair.answer == "Yes":
            n_yes_ans += 1
            n_yes_q_answed_well += 1
        elif pair.answer == "No":
            n_no_q_answed_well += 1
    elif answer:
        if pair.qtype == "F":
            n_fact_q_answed +=1
        elif pair.answer == "Yes":
            n_yes_ans += 1
    else: # no answer case
        n_fact_q_noans += 1

    # Expectations validation:
    if bool(pair.expectation) == valid_answer(answer, pair):
        n_expect_q_valid += 1
        print >> verbose, "Expectation is validated."

    # Recall:
    if correct_answer_in_the_list:
        n_fact_ans_in_list += 1 # Compared with n_fact_q 

        
verbose.close()

print >> sys.stderr, time.strftime('%X'), "- All", len(qas), "Q&As processed."

fact_pre = float(n_fact_q_answed_well) / n_fact_q_answed
yn_pre = float(n_yes_q_answed_well) / n_yes_ans
pre = float(yn_pre * n_yes_ans + fact_pre * n_fact_q_answed) / (n_yes_ans + n_fact_q_answed)


answed = float(n_fact_q_answed) / n_fact_q # only on factoid as is 100% on Y/N
answed_well = float(n_fact_q_answed_well + n_yes_q_answed_well) / (n_fact_q + n_yes_q)

expectations = float(n_expect_q_valid) / n_expect_q

recall = float(n_fact_ans_in_list) / n_fact_q
recall_per_kextrs = recall*100000/num_exts

yn_perf = float(n_yes_q_answed_well + n_no_q_answed_well) / n_yn_q
fact_perf = float(n_fact_q_answed_well) / n_fact_q
perf = float(n_fact_q_answed_well + n_yes_q_answed_well + n_no_q_answed_well) / n_q

avg_num_answ = float(total_num_answers) / n_fact_q
avg_num_answ_when_answd = float(total_num_answers) / n_fact_q_answed

loose_fact_pre = float(n_fact_ans_in_list) / n_fact_q_answed

print >> sys.stderr, time.strftime('%X'), "- Finished evaluation."

system_proper = { "reverb" : "ReVerb", "ollie" : "Ollie", "clausie" : "ClausIE", "all" : "All"}

if args.paper_flag:
    if args.paper_flag == "perf-table":
        # Easy case, print results on a line.
        print "| " + "{:<8}".format(system_proper[args.system]) + "|" + "{:>9}".format("{:.0%}".format(answed)) + " |" + "{:>10}".format("{:.0%}".format(loose_fact_pre)) + " |" + "{:>7}".format("{:.0%}".format(recall))+ " |"
        
    elif args.paper_flag == "parameter-table":
        # Slightly harder case : print results in a column, they'll get `pasted` later.
        print "{:^3}".format("{:.0%}".format(answed))
        print "{:^3}".format("{:.1f}".format(avg_num_answ_when_answd))[:3]
        print "{:^3}".format("{:.0%}".format(loose_fact_pre))
        print "{:^3}".format("{:.0%}".format(recall))
        print "---"
    
    else:
        raise ValueError("What do you mean by passing", args.paper_flag, "as a flag ?")

    sys.exit(0)
    # We finish early not to spoil our ASCII format

print "\n=== Results for the ", args.system, " system ==="

# # Precision of answers to any question: bad metric because of Yes-No questions imbalance.
# print "\nPerformance on", n_yn_q,"Yes-No questions, and", n_fact_q, "factoid questions :", "{:.0%}".format(yn_perf), "-", "{:.0%}".format(fact_perf)
# print "Total performance on", n_q, "questions :\t\t\t\t" , "{:.0%}".format(perf)

print "Precision (tight) :", n_fact_q_answed_well, "answers to", n_fact_q_answed, "factoid questions were correct :\t\t\t", "{:.0%}".format(fact_pre), "(1 answer per question, not well picked ; bad metric)"
print "Precision (loose) :", n_fact_ans_in_list, "/", n_fact_q_answed, "answered factoid questions were given >0 correct answers :\t", "{:.0%}".format(loose_fact_pre), "(among", "{:.1f}".format(avg_num_answ_when_answd), "answers on avg)"

# Recall (good metric)
print "Recall :", n_fact_ans_in_list, "factoid questions out of", n_fact_q, "had >0 correct answers matched down the list :\t", "{:.0%}".format(recall)
print "On", num_exts, "extractions, that's", "{:.1f}".format(recall_per_kextrs) , "recall points per 1k extractions."

print "\n",
# Proportion of answered questions (good to know).
print n_fact_q_noans, "factoid questions out of", n_fact_q, "were left unanswered :\t","{:.0%}".format(1-answed)

# Number of proposed answers per question (good to know).
print "There are","{:.1f}".format(avg_num_answ), "answers per factoid question, or", "{:.1f}".format(avg_num_answ_when_answd), "excluding zeroes."


# Evaluation-script performance metric : did it behave as expected by the annotator ?
print "On", n_expect_q_valid, "out of", n_expect_q, "validating qas the system scored as expected :\t", "{:.0%}".format(expectations)
